"""Merge one or more VCD files"""
import logging
from typing import List, Optional

import pymergevcd.io_manager_interfaces as iomi


class MergeEngine(iomi.AggregatorInterface):
    """Merge multiple files"""

    def __init__(self, sources: List[iomi.AggregatorInterface],
                 datestr: Optional[str] = None):
        self.sources: List[iomi.AggregatorInterface] = sources
        self._datestr = datestr

    # pylint: disable=too-many-branches
    def get_list(self) -> List[iomi.VcdElements]:  # noqa: C901
        """Return elements until end-of-file

        It's a monster of function that extracts all exisiting
        elements. It then proceeds to sort all elements in-memory.

        """
        elements_arr = dict()
        for i in self.sources:
            j = i.get_list()
            logging.info('Length of list: %i', len(j))
            elements_arr[i] = j
            while not isinstance(j[-1], iomi.VcdEndOfFile):
                j = i.get_list()
                logging.info('Length of list: %i', len(j))
                elements_arr[i].extend(j)

        for i in self.sources:
            if not isinstance(elements_arr[i][0], iomi.VcdEndOfFile):
                break
        else:
            return elements_arr[i]

        header: List[iomi.VcdElements] = []
        values: List[iomi.VcdValueChange] = []
        timescale = None
        if self._datestr is not None:
            header.append(iomi.VcdDate(self._datestr))

        for i in self.sources:
            for element in elements_arr[i]:
                if isinstance(element, iomi.VcdValueChange):
                    element.ident = i.namespace() + element.ident
                    values.append(element)
                if isinstance(element, iomi.VcdVariable):
                    if element.scope:
                        element.scope = i.namespace() + '.' + element.scope
                    else:
                        element.scope = i.namespace()
                    logging.info(element.scope)
                    element.ident = i.namespace() + element.ident
                    header.append(element)
                if isinstance(element, iomi.VcdTimescale):
                    assert timescale is None or (
                        element.size == timescale.size
                        and element.units == timescale.units), \
                        "Timescales are different, doesn't merge atm"
                    timescale = element

        assert timescale is not None
        return ([timescale] + header +                      # type: ignore
                sorted(values, key=lambda x: x.timestamp))  # type: ignore

    def namespace(self) -> str:
        """Return namespace of this aggregator"""
        return 'MergeEngine'



# pylint: disable=too-few-public-methods
"""Interface definitions for IO Manager"""
import abc
from typing import List

import vcd


class VcdElements(abc.ABC):
    """Classes to return from parser"""


class AggregatorInterface(abc.ABC):
    """Classes to return from parser"""
    @abc.abstractmethod
    def get_list(self) -> List['VcdElements']:
        """Return elements until end-of-file"""

    @abc.abstractmethod
    def namespace(self) -> str:
        """Return namespace of this aggregator"""


class WriterInterface(abc.ABC):
    """Classes to return from parser"""
    @abc.abstractmethod
    def process_source(self, src: AggregatorInterface):
        """Process an AggregatorInterface source

        Args:
           src: AggregatorInterface source, possibly VcdReaders

    """


class EndOfFile(abc.ABC):
    """The files have finished"""


class VcdEndOfFile(VcdElements, EndOfFile):
    """Signal end-of-file"""


class VcdTimescale(VcdElements):
    """Configured Timescale"""
    def __init__(self, size, units):
        assert 1000 >= int(size) > 0
        assert units in [
            x.value for x in vcd.writer.TimescaleUnit.__members__.values()]
        self.size = size
        self.units = units


class VcdDate(VcdElements):
    """Configured Date"""
    def __init__(self, date):
        self.date = date


class VcdEndOfHeader(VcdElements):
    """Indicate Header Parsing is over, data follows"""


class VcdVariable(VcdElements):
    """Variable"""
    # pylint: disable=too-many-arguments
    def __init__(self, scope, name, var_type, ident, size=None, init=None):
        self.scope = scope
        self.name = name
        self.ident = ident
        self.var_type = var_type
        self.size = size
        self.init = init


class VcdValueChange(VcdElements):
    """New value for `VcdVariable`"""
    def __init__(self, line: list, tstamp: int):
        self.timestamp = tstamp
        if len(line) == 1:
            self.value = line[0][0]
            self.ident = line[0][1:]  # Identifier
        elif len(line) == 2:
            self.value = line[0]
            self.ident = line[1]
        else:
            raise ValueError

        if isinstance(self.value, str) and self.value[0] == 'b':
            if self.value[1] == 'x':
                self.value = self.value[1:]
            else:
                self.value = int(self.value[1:], 2)



"""Parse a VCD file into its elements"""
import logging
from typing import List, Optional

import pymergevcd.io_manager_interfaces as iomi


# pylint: disable=too-few-public-methods
class VcdParserState:
    """Store current parse information and create new `VcdElements`"""
    def __init__(self):
        self.tstamp = 0
        self.register = dict()
        self.scope = []
        self.parsing_values = False

    # pylint: disable=too-many-branches
    def factory(self, line: str) -> Optional[iomi.VcdElements]:  # noqa:C901
        """Convert lines into VcdElements

        Default values are currently not supported. Rather they are
        interpreted as the value at t = 0.

        """
        linel = line.split(' ')
        if line[0] == '#':
            self.tstamp = int(line[1:].strip())
        elif linel[0] == '$timescale':
            return iomi.VcdTimescale(linel[1], linel[2])
        elif linel[0] == '$date':
            return iomi.VcdDate(linel[1])
        elif linel[0] == '$var':
            size = int(linel[2])
            return iomi.VcdVariable('.'.join(self.scope), linel[4], linel[1],
                                    linel[3], size)
        elif linel[0] == '$scope':
            if linel[2]:
                logging.info('Scope: %s %s', linel[2], linel[3])
                self.scope.append(linel[2])
        elif linel[0] == '$upscope':
            if self.scope:
                del self.scope[-1]
            else:
                logging.warning('Scope empty')
            logging.info(self.scope)
        elif linel[0] == '$dumpvars':
            self.parsing_values = True
        elif linel[0] == '$end':
            assert self.parsing_values is True
        elif linel[0] in ['$enddefinitions']:
            return iomi.VcdEndOfHeader()
        elif self.parsing_values:
            return iomi.VcdValueChange(linel, self.tstamp)
        else:
            raise KeyError('Unknown Key in VCD File')
        return None


class VcdReader(iomi.AggregatorInterface):
    """Parse a VCD file into `VcdElements`"""
    def __init__(self, filename: str = ''):
        """Parse given filename"""
        self._filename = filename
        self._elements: List[iomi.VcdElements] = []
        self._is_parsed = False
        self._parse()

    def get_list(self) -> List[iomi.VcdElements]:
        """Return all elements of file"""
        if self._elements:
            cur_el = self._elements[:]
            del self._elements[:len(cur_el)]
            logging.info('Returning %i elements', len(cur_el))
            return cur_el
        if self._is_parsed:
            return [iomi.VcdEndOfFile()]
        return []

    def namespace(self) -> str:
        if self._filename[-4:] == '.vcd':
            namespace = self._filename[:-4]
        namespace = namespace.replace('.', '_')
        return namespace

    def _parse(self):
        logging.info('Starting parsing')
        state = VcdParserState()
        for line in open(self._filename):
            parsed_line = state.factory(line.strip())
            if parsed_line:
                self._elements.append(parsed_line)
        self._elements.append(iomi.VcdEndOfFile())
        self._is_parsed = True
        logging.info('Terminating parsing: %i elements', len(self._elements))


def factory(filename: str) -> VcdReader:
    """Create an `AggregatorInterface` reader depending on filename"""
    if filename.endswith('.vcd'):
        return VcdReader(filename)
    raise ValueError('Unexpected file-ending')



#!/usr/bin/env python3

from pathlib import Path

import setuptools

project_dir = Path(__file__).parent

setuptools.setup(
    name='pymergevcd',
    version='0.6',

    description='Merge various VCD files',

    # Allow UTF-8 characters in README with encoding argument.
    long_description=project_dir.joinpath('README.md').read_text(encoding='utf-8'),
    long_description_content_type="text/markdown",
    keywords=['python'],

    author='Kristoffer Nordström',
    url='https://github.com/kown7/pymergevcd',

    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},

    # pip 9.0+ will inspect this field when installing to help users install a
    # compatible version of the library for their Python version.
    python_requires='>=3.6',

    # There are some peculiarities on how to include package data for source
    # distributions using setuptools. You also need to add entries for package
    # data to MANIFEST.in.
    # See https://stackoverflow.com/questions/7522250/
    include_package_data=True,

    # This is a trick to avoid duplicating dependencies between both setup.py and
    # requirements.txt.
    # requirements.txt must be included in MANIFEST.in for this to work.
    # It does not work for all types of dependencies (e.g. VCS dependencies).
    # For VCS dependencies, use pip >= 19 and the PEP 508 syntax.
    #   Example: 'requests @ git+https://github.com/requests/requests.git@branch_or_tag'
    #   See: https://github.com/pypa/pip/issues/6162
    install_requires=project_dir.joinpath('requirements.txt').read_text().split('\n'),
    zip_safe=False,

    license='MIT',
    license_files=['LICENSE.txt'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

    entry_points={
        'console_scripts': ['pymergevcd=pymergevcd.cli:main'],
    }
)
